// Copyright (c) 2024 Tulir Asokan
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

package hicli

import (
	"context"
	"fmt"
	"net/url"

	"maunium.net/go/mautrix"
	"maunium.net/go/mautrix/hicli/database"
	"maunium.net/go/mautrix/id"
)

var InitialDeviceDisplayName = "mautrix hiclient"

func (h *HiClient) LoginPassword(ctx context.Context, homeserverURL, username, password string) error {
	var err error
	h.Client.HomeserverURL, err = url.Parse(homeserverURL)
	if err != nil {
		return err
	}
	return h.Login(ctx, &mautrix.ReqLogin{
		Type: mautrix.AuthTypePassword,
		Identifier: mautrix.UserIdentifier{
			Type: mautrix.IdentifierTypeUser,
			User: username,
		},
		Password:                 password,
		InitialDeviceDisplayName: InitialDeviceDisplayName,
	})
}

func (h *HiClient) Login(ctx context.Context, req *mautrix.ReqLogin) error {
	err := h.CheckServerVersions(ctx)
	if err != nil {
		return err
	}
	req.StoreCredentials = true
	req.StoreHomeserverURL = true
	resp, err := h.Client.Login(ctx, req)
	if err != nil {
		return err
	}
	h.Account = &database.Account{
		UserID:        resp.UserID,
		DeviceID:      resp.DeviceID,
		AccessToken:   resp.AccessToken,
		HomeserverURL: h.Client.HomeserverURL.String(),
	}
	h.CryptoStore.AccountID = resp.UserID.String()
	h.CryptoStore.DeviceID = resp.DeviceID
	err = h.DB.Account.Put(ctx, h.Account)
	if err != nil {
		return err
	}
	err = h.Crypto.Load(ctx)
	if err != nil {
		return fmt.Errorf("failed to load olm machine: %w", err)
	}
	err = h.Crypto.ShareKeys(ctx, 0)
	if err != nil {
		return err
	}
	_, err = h.Crypto.FetchKeys(ctx, []id.UserID{h.Account.UserID}, true)
	if err != nil {
		return fmt.Errorf("failed to fetch own devices: %w", err)
	}
	return nil
}

func (h *HiClient) LoginAndVerify(ctx context.Context, homeserverURL, username, password, recoveryCode string) error {
	err := h.LoginPassword(ctx, homeserverURL, username, password)
	if err != nil {
		return err
	}
	err = h.VerifyWithRecoveryCode(ctx, recoveryCode)
	if err != nil {
		return err
	}
	go h.Sync()
	go h.RunRequestQueue(ctx)
	return nil
}
